// Copyright (c) 2021, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
// All rights reserved.
//
// This file is licensed according to the BSD 3-clause license as follows:
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the "Linz Center of Mechatronics GmbH" and "LCM" nor
//       the names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL "Linz Center of Mechatronics GmbH" BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// This file is part of X2C. http://x2c.lcm.at/
// $LastChangedRevision: 1603 $
// $LastChangedDate:: 2019-01-21 19:02:13 +0100#$
//
// This function returns the URL to the block image suitable for the Xcos JGraphX style property.
// It returns an empty string if the image file does not exist.
// 
// If the 'image_name' parameter is specified, the function tries to use this altivernative block image file.
// The file name + suffix are extracted from 'image_name', the path is discarded.
// This allows to use one (or multiple) other image file(s) by one block.
// If no suffix is present, the function adds the default suffix '.svg'.
// 
// The 'kill_dummy' parameter allows to detect the usage of the dummy image file when an alternative block image is provided.
// If set to true, the function returns an empty string if usage of the dummy file is detected.
// 
// Example: image_url = getStyleImageLink("General", "Gain");
//          image_url = getStyleImageLink("General", "Gain", "MyGain0.svg");
//          image_url = getStyleImageLink("General", "Gain", "C:\MyImageDir\MyGain0.svg");
//          image_url = getStyleImageLink("General", "Gain", "C:\MyImageDir\Dummy.svg", %t);
// 
// Parameters:
// library_name ... X2C library name
// block_type ... X2C library type name
// image_name (optional) ... location or file name of an alternative block image file
// kill_dummy (optional) ... true to enable detection of usage of the dummy image file
// 
// Return values:
// image_url ... Block image URL
//               Empty string if block image does not exist
//               Empty string if dummy detection is enabled and a dummy image file was found

function [image_url] = getStyleImageLink(library_name, block_type, image_name, kill_dummy)
    funcprot(0);
    
    default_suffix = ".svg";

    if isdef('image_name', 'l')
        // strip path information and adds default suffix, if source file does not contain a suffix
        [path, name, suffix] = fileparts(image_name);
        if ~isdef('kill_dummy', 'l')
            kill_dummy = %f;
        end
        
        // return empty string if dummy kill is enabled and the file name equals 'dummy' (case insensitive)
        if kill_dummy & (strcmp(name, 'dummy', 'i') == 0)
            image_url = "";
            return;
        end
        
        if isempty(suffix)
            image_filename = name + default_suffix;
        else
            image_filename = name + suffix;
        end
    else
        image_filename = block_type + default_suffix;
    end

    jimport("at.lcm.x2c.core.librarymanager.X2cLibraryManager");
    lib_root = jinvoke(jinvoke(jinvoke(jinvoke(X2cLibraryManager, "getInstance"), "getLibrary", library_name), "getLibraryPath"), "getAbsolutePath");
    new_img_location = lib_root + filesep() + "Doc" + filesep() + "Images" + filesep() + image_filename;
    
    // returns link only if file exists
    if isfile(new_img_location)
        // add URL protocol prefix and replace directory separators by slash character
        image_url = "file:///" + strsubst(new_img_location, filesep(), "/");   
    else
        image_url = "";
    end 
endfunction
